#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) KeyChest Ltd, registered in the United Kingdom.
 This file is owned exclusively by KeyChest Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


try:
    import pkg_resources
    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    # noinspection PyUnboundLocalVariable
    __path__ = pkgutil.extend_path(__path__, __name__)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***
"""
#  Copyright (c) KeyChest Ltd, 2019. File provided under the license as described in the "LICENSE" file
#   included with this project. A copy of the license is also available from https://gitlab.com/keychest,
#   when you select the relevant project or you can request it at support@keychest.net.
#
#   This file is owned exclusively by KeyChest Ltd. Unauthorized copying of this file, via any medium is
#    strictly prohibited
#
import base64
import json
import logging
import time
import traceback
from urllib.parse import urlparse

__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


class AcmeLog(object):

    @staticmethod
    def repeater(req_hdrs, code, path, req_body, resp_body, resp_hdrs):
        """
        This is a simple method that checks messages, parses selected ones and logs them
        :param resp_body:
        :type code: int
        :param code: response code
        :param req_hdrs: headers of the request - it includes 'Host'/server we assume prod/staging for LE but it can be any
        :param path: path of the GET/POST request
        :param req_body: data sent from the client
        :return:
        """
        try:
            client = req_hdrs['X-Forwarded-For']
            server = req_hdrs['Host']
            agent = req_hdrs['User-Agent']
            if resp_hdrs:
                location = resp_hdrs['Location']
            else:
                location = None

            body = AcmeLog.decode_acme2(req_body)
            response = AcmeLog.decode_response(resp_body)

            logging.debug("repeater call", path=path, code=code, server=server, client=client, agent=agent)
            _path = path.lower()
            if '/directory' in path:
                logging.usage("directory", epoch=int(time.time()), client=client, server=server, code=code, agent=agent)
                pass
            elif 'acme/new-acct' in path:
                if ('contact' not in response) or ('initialIp' not in response) or ('status' not in response):
                    logging.error("Unrecognized body of new-acct", msg=req_body)
                elif location is None:
                    logging.error("Missing header Location in new-acct", msg=req_body, resp=resp_body)
                else:
                    email = "acme@keychest.net"
                    for each_contact in response['contact']:  # type: str
                        if each_contact.startswith('mailto:'):
                            email = each_contact.split(":")[1].strip().lower()

                    account = urlparse(location).path

                    logging.usage("new-acct", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  contact=response['contact'], email=email, account=account, source=response['initialIp'],
                                  status=response['status'])
                    pass
            elif 'acme/new-nonce' in path:
                logging.usage("new-nonce", epoch=int(time.time()), client=client, server=server, code=code, agent=agent)
                pass
            elif 'acme/new-order' in path:
                if (body is None) or ('kid' not in body):
                    logging.error("Unrecognized body of new-order", msg=req_body)
                elif ('identifiers' not in response) or ('authorizations' not in response) or ('finalize' not in response):
                    logging.error("Unrecognized response to new-order", msg=response)
                elif location is None:
                    logging.error("Missing location header in new-order", msg=response)
                else:
                    authz = []
                    for x in response['authorizations']:
                        _x = urlparse(x)
                        authz.append(_x.path)
                    finalize = urlparse(response['finalize']).path
                    kid = urlparse(body['kid']).path
                    loc = urlparse(location).path
                    identifiers = response['identifiers']
                    logging.usage("new-order", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  domains=identifiers, authz=authz, finalize=finalize, account=kid, order=loc)
                pass
            elif 'acme/revoke-cert' in path:
                if (body is None) or ('kid' not in body):
                    logging.error("Unrecognized body of revoke-cert", msg=req_body)
                else:
                    kid = urlparse(body['kid']).path
                    logging.usage("revoke-cert", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid)
            elif 'acme/key-change' in path:
                if (body is None) or ('kid' not in body):
                    logging.error("Unrecognized body of key-change", msg=req_body)
                else:
                    kid = urlparse(body['kid']).path
                    logging.usage("key-change", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid)
                pass
            elif 'acme/auth' in path:
                if (body is None) or ('kid' not in body) or ('url' not in body):
                    logging.error("Unrecognized body of auth", msg=req_body)
                elif (response is None) or ('status' not in response) or ('expires' not in response) or ('identifier' not in response):
                    logging.error("Unrecognized body of auth response", msg=resp_body)
                else:
                    kid = urlparse(body['kid']).path
                    authz = urlparse(body['url']).path
                    status = response['status']
                    expires = response['expires']
                    identifier = response['identifier']
                    logging.usage("auth", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid, authz=authz, status=status, expires=expires, identifier=identifier)
                pass
            elif 'acme/accnt' in path:
                pass
            elif 'acme/final' in path:
                if (body is None) or ('kid' not in body) or ('url' not in body):
                    logging.error("Unrecognized body of final", msg=req_body)
                elif (response is None) or ('certificate' not in response) or ('status' not in response):
                    logging.error("Unrecognized response of final", msg=resp_body)
                else:
                    kid = urlparse(body['kid']).path
                    loc = urlparse(body['url']).path
                    status = response['status']
                    cert = response['certificate']
                    cert = urlparse(cert).path
                    logging.usage("final", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid, status=status, cert=cert)
                pass
            elif 'acme/chall' in path:
                if (body is None) or ('kid' not in body) or ('url' not in body):
                    logging.error("Unrecognized body of chall", msg=req_body)
                elif (response is None) or ('status' not in response):
                    logging.error("Unrecognized response of chall", msg=resp_body)
                else:
                    kid = urlparse(body['kid']).path
                    loc = urlparse(body['url']).path
                    status = response['status']
                    logging.usage("chall", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid, chall=loc, status=status)
                pass
            elif 'acme/order' in path:
                if (body is None) or ('kid' not in body) or ('url' not in body):
                    logging.error("Unrecognized body of order", msg=req_body)
                elif (response is None) or ('identifiers' not in response) or ('status' not in response) \
                        or ('authorizations' not in response) or ('finalize' not in response):
                    logging.error("Unrecognized response of order", msg=resp_body)
                else:
                    kid = urlparse(body['kid']).path
                    loc = urlparse(body['url']).path
                    status = response['status']
                    identifiers = response['identifiers']
                    authorizations = response['authorizations']
                    finalize = response['finalize']
                    if 'certificate' in response:
                        certificate = response['certificate']
                    else:
                        certificate = ""
                    authz = []
                    for _x in authorizations:
                        authz.append(urlparse(_x).path)
                    final = urlparse(finalize).path
                    logging.usage("order", epoch=int(time.time()), client=client, server=server, code=code, agent=agent,
                                  account=kid, chall=loc, status=status, identifiers=identifiers, finalize=final, certificate=certificate)
            else:
                logging.error("Unrecognized message", path=path)
        except Exception as ex:
            tb = "; ".join(traceback.format_exc(3).splitlines())
            logging.error("Unexpected exception in repeater", cause=str(ex), traceback=tb)
        return

    @staticmethod
    def decode_acme2(msg):

        if (msg is None) or len(msg) < 2:
            return None
        try:
            tamper = json.loads(msg)

            if 'protected' in tamper:
                protected = tamper['protected']
                protected_plain = base64.b64decode(protected + "=" * ((4 - len(protected) % 4) % 4))
                protected_json = json.loads(protected_plain.decode())
                return protected_json
            else:
                return tamper
        except Exception as ex:
            logging.info("Error parsing ACMEv2 message", msg=msg.decode(), cause=str(ex))
            return None

    @staticmethod
    def decode_response(msg):

        if (msg is None) or len(msg) < 2:
            return None
        try:
            j_msg = json.loads(msg)
            return j_msg
        except Exception as ex:
            logging.info("Error parsing ACMEv2 response", msg=msg.decode(), cause=str(ex))
            return None



# -*- coding: utf-8 -*-
# Copyright (c) 2015, inaz2
# All rights reserved. BSD 3-Clause "New" or "Revised" License
# https://github.com/inaz2/proxy2
# Copyright (c) 2019, KeyChest Ltd  support@keychest.net
# updated for Python3, logging replaced with async logging using logbook, smaller changes throughout
import logging
import sys
import socket
import ssl
import select
import http.client as httplib
from urllib.parse import urlsplit
from urllib.parse import parse_qsl as urllib_qsl
from urllib.parse import urlparse as urlparse

import threading
import gzip
import zlib
import json
import re
from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
from io import StringIO
from html.parser import HTMLParser

from keychestamp.acme_log import AcmeLog
from keychestamp.ampconfig import AmpConfig

COLOR_ON = False


def with_color(c, s):
    if COLOR_ON:
        return "\x1b[%dm%s\x1b[0m" % (c, s)
    else:
        return s


class ThreadingHTTPServer(ThreadingMixIn, HTTPServer):
    address_family = socket.AF_INET6
    daemon_threads = True

    def handle_error(self, request, client_address):
        # suppress socket/ssl related errors
        cls, e = sys.exc_info()[:2]
        if cls is socket.error or cls is ssl.SSLError:
            pass
        else:
            return HTTPServer.handle_error(self, request, client_address)


class ProxyRequestHandler(BaseHTTPRequestHandler):

    timeout = AmpConfig.api_timeout
    lock = threading.Lock()

    def __init__(self, *args, **kwargs):
        self.tls = threading.local()
        self.tls.conns = {}
        self.connection = None
        self.rfile = None
        self.wfile = None
        self.close_connection = None

        BaseHTTPRequestHandler.__init__(self, *args, **kwargs)

    def log_error(self, _format, *args):
        # suppress "Request timed out: timeout('timed out',)"
        if isinstance(args[0], socket.timeout):
            return
        self.log_message(_format, *args)

    def do_CONNECT(self):
        if AmpConfig.cert_ready():
            self.connect_intercept()
        else:
            logging.error("Proxy key or certificate are not ready")
            self.connect_relay()

    def connect_intercept(self):
        hostname = self.path.split(':')[0]

        with self.lock:
            AmpConfig.ensure_name_cert(hostname)

        message = "%s %d %s\r\n\r\n" % (self.protocol_version, 200, 'Connection Established')

        self.wfile.write(message.encode())

        self.connection = ssl.wrap_socket(self.connection, keyfile=AmpConfig.cert_key, certfile=AmpConfig.get_cert_path(hostname),
                                          server_side=True)
        self.rfile = self.connection.makefile("rb", self.rbufsize)
        self.wfile = self.connection.makefile("wb", self.wbufsize)

        conntype = self.headers.get('Proxy-Connection', '')
        if self.protocol_version == "HTTP/1.1" and conntype.lower() != 'close':
            self.close_connection = 0
        else:
            self.close_connection = 1

    def connect_relay(self):
        address = self.path.split(':', 1)
        address[1] = int(address[1]) or 443
        try:
            s = socket.create_connection(address, timeout=self.timeout)
        except Exception as e:
            logging.info("Relay error", cause=str(e))
            self.send_error(502)
            return
        self.send_response(200, 'Connection Established')
        self.end_headers()

        conns = [self.connection, s]
        self.close_connection = 0
        while not self.close_connection:
            rlist, wlist, xlist = select.select(conns, [], conns, self.timeout)
            if xlist or not rlist:
                break
            for r in rlist:
                other = conns[1] if r is conns[0] else conns[0]
                data = r.recv(8192)
                if not data:
                    self.close_connection = 1
                    break
                other.sendall(data)

    def do_GET(self):
        if self.path == 'http://amp.keychest.net/':
            self.send_ca_cert()
            return

        req = self
        content_length = int(req.headers.get('Content-Length', 0))
        req_body = self.rfile.read(content_length) if content_length else None

        # req.path = "http://127.0.0.1:6443" + req.path
        if req.path[0] == '/':
            if isinstance(self.connection, ssl.SSLSocket):
                req.path = "https://%s%s" % (req.headers['Host'], req.path)
            else:
                req.path = "http://%s%s" % (req.headers['Host'], req.path)

        # noinspection PyNoneFunctionAssignment
        req_body_modified = self.request_handler(req, req_body)
        if req_body_modified is False:
            self.send_error(403)
            return
        elif req_body_modified is not None:
            req_body = req_body_modified
            # noinspection PyTypeChecker
            req.headers['Content-length'] = str(len(req_body))

        req.headers['X-Forwarded-For'] = req.client_address[0]

        u = urlparse(req.path)
        scheme, netloc, path = u.scheme, u.netloc, (u.path + '?' + u.query if u.query else u.path)
        assert scheme in ('http', 'https')
        if netloc:
            req.headers['Host'] = netloc
        setattr(req, 'headers', self.filter_headers(req.headers))

        origin = (scheme, netloc)
        res_body = None
        res_code = None
        res = None
        try:
            if origin not in self.tls.conns:
                if scheme == 'https':
                    if (ssl.get_default_verify_paths().cafile is None) and (ssl.get_default_verify_paths().capath is None):
                        logging.error("No root CA bundle configured for Python")
                        # noinspection PyProtectedMember
                        context = ssl._create_unverified_context()
                    else:
                        context = ssl.create_default_context()
                    self.tls.conns[origin] = httplib.HTTPSConnection(netloc, timeout=self.timeout, context=context)
                else:
                    self.tls.conns[origin] = httplib.HTTPConnection(netloc, timeout=self.timeout)
            conn = self.tls.conns[origin]
            conn.request(self.command, path, req_body, dict(req.headers))
            res = conn.getresponse()
            res_code = res.code

            version_table = {10: 'HTTP/1.0', 11: 'HTTP/1.1'}
            setattr(res, 'headers', res.msg)
            setattr(res, 'response_version', version_table[res.version])

            # support streaming
            if ('Content-Length' not in res.headers) and ('no-store' in res.getheader('Cache-Control', '')):
                self.response_handler(req, req_body, res, '')
                setattr(res, 'headers', self.filter_headers(res.headers))
                self.relay_streaming(res)
                with self.lock:
                    self.save_handler(req, req_body, res, '')
                return

            res_body = res.read()
        except Exception as e:
            logging.info("Get request error", cause=str(e))
            if origin in self.tls.conns:
                del self.tls.conns[origin]
            self.send_error(502)
            return
        finally:
            res_hdrs = None
            if res and res.headers:
                res_hdrs = res.headers
            AcmeLog.repeater(req.headers, res_code, req.path, req_body, res_body, res_hdrs)

        content_encoding = res.getheader('Content-Encoding', 'identity')
        res_body_plain = self.decode_content_body(res_body, content_encoding)

        # noinspection PyNoneFunctionAssignment
        res_body_modified = self.response_handler(req, req_body, res, res_body_plain)
        if res_body_modified is False:
            self.send_error(403)
            return
        elif res_body_modified is not None:
            res_body_plain = res_body_modified
            res_body = self.encode_content_body(res_body_plain, content_encoding)
            res.headers['Content-Length'] = str(len(res_body))

        setattr(res, 'headers', self.filter_headers(res.headers))

        message = "%s %d %s\r\n" % (self.protocol_version, res.status, res.reason)
        self.wfile.write(message.encode())
        for line in res.getheaders():
            self.send_header(line[0], line[1])
        self.end_headers()
        if res_body and len(res_body) > 0:
            self.wfile.write(res_body)
        self.wfile.flush()

        with self.lock:
            self.save_handler(req, req_body, res, res_body_plain)

    def relay_streaming(self, res):
        message = "%s %d %s\r\n" % (self.protocol_version, res.status, res.reason)
        self.wfile.write(message.encode())
        for line in res.getheaders():
            self.send_header(line[0], line[1])
        self.end_headers()
        try:
            while True:
                chunk = res.read(8192)
                if not chunk:
                    break
                self.wfile.write(chunk)
            self.wfile.flush()
        except socket.error:
            # connection closed by client
            pass

    do_HEAD = do_GET
    do_POST = do_GET
    do_PUT = do_GET
    do_DELETE = do_GET
    do_OPTIONS = do_GET

    @staticmethod
    def filter_headers(headers):
        # http://tools.ietf.org/html/rfc2616#section-13.5.1
        hop_by_hop = ('connection', 'keep-alive', 'proxy-authenticate', 'proxy-authorization', 'te', 'trailers',
                      'transfer-encoding', 'upgrade')
        for k in hop_by_hop:
            del headers[k]

        # accept only supported encodings
        if 'Accept-Encoding' in headers:
            ae = headers['Accept-Encoding']
            filtered_encodings = [x for x in re.split(r',\s*', ae) if x in ('identity', 'gzip', 'x-gzip', 'deflate')]
            headers['Accept-Encoding'] = ', '.join(filtered_encodings)

        return headers

    @staticmethod
    def encode_content_body(text, encoding):
        if encoding == 'identity':
            data = text
        elif encoding in ('gzip', 'x-gzip'):
            io = StringIO()
            with gzip.GzipFile(fileobj=io, mode='wb') as f:
                f.write(text)
            data = io.getvalue()
        elif encoding == 'deflate':
            data = zlib.compress(text)
        else:
            raise Exception("Unknown Content-Encoding: %s" % encoding)
        return data

    @staticmethod
    def decode_content_body(data, encoding):
        if encoding == 'identity':
            text = data
        elif encoding in ('gzip', 'x-gzip'):
            io = StringIO(data)
            with gzip.GzipFile(fileobj=io) as f:
                text = f.read()
        elif encoding == 'deflate':
            try:
                text = zlib.decompress(data)
            except zlib.error:
                text = zlib.decompress(data, -zlib.MAX_WBITS)
        else:
            raise Exception("Unknown Content-Encoding: %s" % encoding)
        return text

    def send_ca_cert(self):
        with open(AmpConfig.ca_cert, 'rb') as f:
            data = f.read()

        first_line = "%s %d %s\r\n" % (self.protocol_version, 200, 'OK')
        self.wfile.write(first_line.encode())
        self.send_header('Content-Type', 'application/x-x509-ca-cert')
        self.send_header('Content-Length', len(data))
        self.send_header('Connection', 'close')
        self.end_headers()
        self.wfile.write(data)

    @staticmethod
    def print_info(req, req_body, res, res_body):
        def parse_qsl(s):
            return '\n'.join("%-20s %s" % (k, v) for k, v in urllib_qsl(s, keep_blank_values=True))

        req_header_text = "%s %s %s\n%s" % (req.command, req.path, req.request_version, req.headers)
        res_header_text = "%s %d %s\n%s" % (res.response_version, res.status, res.reason, res.headers)

        logging.debug(with_color(33, req_header_text))

        u = urlsplit(req.path)
        if u.query:
            query_text = parse_qsl(u.query)
            logging.debug(with_color(32, "==== QUERY PARAMETERS ====\n%s\n" % query_text))

        cookie = req.headers.get('Cookie', '')
        if cookie:
            cookie = parse_qsl(re.sub(r';\s*', '&', cookie))
            logging.debug(with_color(32, "==== COOKIE ====\n%s\n" % cookie))

        auth = req.headers.get('Authorization', '')
        if auth.lower().startswith('basic'):
            token = auth.split()[1].decode('base64')
            logging.debug(with_color(31, "==== BASIC AUTH ====\n%s\n" % token))

        if req_body is not None:
            req_body_text = None
            content_type = req.headers.get('Content-Type', '')

            if content_type.startswith('application/x-www-form-urlencoded'):
                req_body_text = parse_qsl(req_body)
            elif content_type.startswith('application/json'):
                try:
                    json_obj = json.loads(req_body)
                    json_str = json.dumps(json_obj, indent=2)
                    if json_str.count('\n') < 50:
                        req_body_text = json_str
                    else:
                        lines = json_str.splitlines()
                        req_body_text = "%s\n(%d lines)" % ('\n'.join(lines[:50]), len(lines))
                except ValueError:
                    req_body_text = req_body
            elif len(req_body) < 1024:
                req_body_text = req_body

            if req_body_text:
                logging.debug(with_color(32, "==== REQUEST BODY ====\n%s\n" % req_body_text))

        logging.debug(with_color(36, res_header_text))

        cookies = res.getheader('Set-Cookie')
        if cookies:
            cookies = '\n'.join(cookies)
            logging.debug(with_color(31, "==== SET-COOKIE ====\n%s\n" % cookies))

        if res_body is not None:
            res_body_text = None
            content_type = res.getheader('Content-Type', "")

            if content_type.startswith('application/json'):
                try:
                    json_obj = json.loads(res_body)
                    json_str = json.dumps(json_obj, indent=2)
                    if json_str.count('\n') < 50:
                        res_body_text = json_str
                    else:
                        lines = json_str.splitlines()
                        res_body_text = "%s\n(%d lines)" % ('\n'.join(lines[:50]), len(lines))
                except ValueError:
                    res_body_text = res_body
            elif content_type.startswith('text/html'):
                m = re.search(r'<title[^>]*>\s*([^<]+?)\s*</title>', res_body.decode(), re.I)
                if m:
                    # noinspection PyUnusedLocal
                    h = HTMLParser()
                    # print(with_color(32, "==== HTML TITLE ====\n%s\n" % h.unescape(m.group(1).decode('utf-8'))))
            elif content_type.startswith('text/') and len(res_body) < 1024:
                res_body_text = res_body.decode()

            if res_body_text:
                logging.debug(with_color(32, "==== RESPONSE BODY ====\n%s\n" % res_body_text))

    def request_handler(self, req, req_body):
        pass

    def response_handler(self, req, req_body, res, res_body):
        pass

    def save_handler(self, req, req_body, res, res_body):
        self.print_info(req, req_body, res, res_body)


# noinspection PyPep8Naming
def test(HandlerClass=ProxyRequestHandler, ServerClass=ThreadingHTTPServer, protocol="HTTP/1.1"):
    if sys.argv[1:]:
        port = int(sys.argv[1])
    else:
        port = 8080
    server_address = ('127.0.0.1', port)

    HandlerClass.protocol_version = protocol
    httpd = ServerClass(server_address, HandlerClass)

    sa = httpd.socket.getsockname()
    logging.info("Serving HTTP Proxy on", iface=sa[0], port=sa[1])
    httpd.serve_forever()


if __name__ == '__main__':
    logging.debug("keychestamp proxy")
    test()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) KeyChest Ltd, registered in the United Kingdom.
 This file is owned exclusively by KeyChest Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import logging
import os
import signal
import sys
import time
import threading

from keychestamp import ampconfig
from keychestamp.ampconfig import AmpConfig
from keychestamp.proxy import ProxyRequestHandler, ThreadingHTTPServer

__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


# noinspection PyUnusedLocal
def signal_handler(sig, frame):
    """
    Signal handler
    :param sig:
    :param frame:
    :return:
    """

    AmpConfig.main_stop.set()
    AmpConfig.ch.close()
    sys.stderr.write("Waiting 1 second to exit\n")
    logger = logging.getLogger()
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)  # to avoid spurious error messages

    time.sleep(1)

    raise SystemExit('Terminating KeychestAmp')  # this is supposed to kill main()


def process_watch():
    """
    A simple process watch function that will try to close worker processes that died
    :return:
    """

    counter = 1
    # noinspection PyTypeChecker
    watched_thread = None  # type: threading.Thread
    logging.debug("Proxy watcher started at %d" % int(time.time()))
    while not AmpConfig.main_stop.is_set():
        # noinspection PyBroadException
        try:

            if (watched_thread is None) or (not watched_thread.is_alive()):
                watched_thread = threading.Thread(target=process_proxy, args=())
                watched_thread.name = "proxy_thread_" + str(counter)
                watched_thread.daemon = True
                counter += 1
                watched_thread.start()

            AmpConfig.main_stop.wait(timeout=1.0)
        except Exception as e:
            logging.error("Exception in the proxy watcher: %s" % str(e))
            pass
    sys.stderr.write("Process watcher terminates\n")
    sys.stderr.flush()


def process_proxy():

    server_address = (AmpConfig.address, AmpConfig.port)

    ProxyRequestHandler.protocol_version = "HTTP/1.1"
    logging.debug("Starting the proxy HTTP server", address=AmpConfig.address, port=AmpConfig.port)
    httpd = ThreadingHTTPServer(server_address, ProxyRequestHandler)

    sa = httpd.socket.getsockname()
    logging.info("Serving HTTP Proxy on", iface=sa[0], port=sa[1])
    httpd.serve_forever()
    pass


def main():
    """
    Main entry function
    """

    sys.stderr.write("KeyChestAmp started\n")
    # noinspection PyBroadException
    try:
        signal.signal(signal.SIGINT, signal_handler)
    except Exception:
        pass
    # noinspection PyBroadException
    try:
        signal.signal(signal.SIGTERM, signal_handler)
    except Exception:
        pass

    to_exit = ampconfig.AmpConfig.parse_params()
    if to_exit < 0:
        sys.stderr.write("Initialization failed, more information should be above and/or in log files in /var/log/keychestamp/\n\n")
        sys.exit(-1)
    elif to_exit > 0:
        sys.exit(0)

    AmpConfig.set_logging()
    logging.debug("KeyChestAmp launching the proxy")

    # check keys and create if not available
    AmpConfig.ensure_name_cert('acme-staging-v02.api.letsencrypt.org')
    AmpConfig.ensure_name_cert('acme-v02.api.letsencrypt.org')

    # as we use main_stop as multiprocessing.Event there must not be wait() in the main thread.
    # otherwise the is_set() and wait() deadlock each other
    watch_thread = threading.Thread(target=process_watch, args=())
    watch_thread.daemon = True
    watch_thread.name = "Process Watch"
    watch_thread.start() # start the watcher, which will keep the proxy running
    # the above is blocking - only get past when terminating
    while watch_thread.is_alive():
        time.sleep(1)

    sys.stderr.write("KeyChestAmp terminated - main\n")
    sys.stderr.flush()

    # let's launch the watcher thread

    # alternative logbook logging
    # log_dispatcher_thread = LogDispatcher(logger_stop_event, ip="127.0.0.1")
    # log_dispatcher_thread.name = "LogDispatcher"
    # log_dispatcher_thread.start()

    # setup_logging(log_level, true_socket=True)  # alternative logbook logging
    # logging.getLogger().setLevel(log_level)


if __name__ == '__main__':
    # !!!! just for testing - not executed once installed
    try:
        main()
    except Exception as ex:
        sys.stderr.write(str(ex))
        sys.stderr.flush()
    finally:
        sys.stderr.write("KeyChestAmp terminated\n")
        sys.stderr.flush()
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***
"""
#  Copyright (c) KeyChest Ltd, 2019. File provided under the license as described in the "LICENSE" file
#   included with this project. A copy of the license is also available from https://gitlab.com/keychest,
#   when you select the relevant project or you can request it at support@keychest.net.
#
#   This file is owned exclusively by KeyChest Ltd. Unauthorized copying of this file, via any medium is
#    strictly prohibited
#
import errno
import getopt
import json
import logging
import multiprocessing
import os
import random
import socket
import string
import sys
import time
from subprocess import Popen, PIPE

from keychestamp.queue_handler import QueueHandler
from keychestamp.version import proxy_version

__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


class AmpConfig(object):
    """
    A static class with the configuration of the process.
    """

    port = 8443
    address = "::"
    api_logging = True
    file_logging = None
    log_level = logging.INFO
    api_timeout = 5
    ch = None
    id = "dummy@amp.keychest.net"
    temp_ca = None
    main_stop = multiprocessing.Event()

    # noinspection PyBroadException
    try:
        CONFIG_DIR = os.environ['KC_CONFIG_DIR']
    except Exception:
        # hopefully just the key is not set
        CONFIG_DIR = os.path.expanduser("~/.keychestamp")
        # noinspection PyUnusedLocal,PyBroadException
        try:
            _path = os.path.join(CONFIG_DIR)
        except Exception as ex:
            CONFIG_DIR = os.path.expanduser("~")

    config_dir = CONFIG_DIR

    ca_dir = None
    ca_key = None
    ca_cert = None
    cert_key = None
    cert_dir = None
    config = None

    @staticmethod
    def print_help():
        """
        A simple help printing
        """
        print("keychestamp, version %s. \n(c) 2019 KeyChest Ltd. Visit https://gitlab.com/keychest .\n"
              "The proxy accepts the following parameters:" % proxy_version)
        print("  -h, --help - this help, if present it will stop the process")
        print("  -l, --level <DEBUG|ERROR|WARNING|INFO|CRITICAL> - minimum level to log, default is INFO")
        print("  -n, --noapi - disables API logging, starts logging into /var/log/keychestamp/audit.json")
        print("  -f --file <file> - enable local login and set a custom location")
        print("  -p, --port <int> - setting the listening port of the proxy")
        print("  -a, --addr <IP addr> - by default it responds to all requests")
        print("  -e, --etc <folder> - working directory for storing config. data")
        print("  -c, --ca - just returns its root certificate in PEM format and exits\n\n")

    @staticmethod
    def set_ca_env():
        AmpConfig.ca_dir = os.path.join(AmpConfig.config_dir, 'ca')
        AmpConfig.ca_key = os.path.join(AmpConfig.ca_dir, 'private/ca.key')
        AmpConfig.ca_cert = os.path.join(AmpConfig.ca_dir, 'certs/ca.crt')
        AmpConfig.cert_key = os.path.join(AmpConfig.ca_dir, 'private/cert.key')
        AmpConfig.cert_dir = os.path.join(AmpConfig.ca_dir, 'certs/')
        AmpConfig.config = os.path.join(AmpConfig.ca_dir, 'openssl.conf')

    @staticmethod
    def _test_writable(_path):

        # first test it exists, if not, we will create it
        try:
            if not os.path.exists(_path):
                os.makedirs(_path)

            # test if we can read and write
            _file = os.path.join(_path, ".test_file")

            with open(_file, "w") as f:
                x = str(time.time())
                f.write(x)
            with open(_file, "r") as f:
                x2 = f.read()
            if x != x2:
                sys.stderr.write("Working directory not writeable\n")
                return -1
            os.remove(_file)
        except Exception as ex:
            sys.stderr.write("Can't write into the working directory: %s\n" % str(ex))
            return -1

        _dir = os.path.join(_path, '.test_dir')
        try:
            if os.path.exists(_dir):
                os.rmdir(_dir)
            os.mkdir(_dir, 0x770)
            os.rmdir(_dir)
        except Exception as ex:
            sys.stderr.write("Can't create folder in the working directory: %s\n" % str(ex))
            return -1
        return 0

    @staticmethod
    def _test_config(path):

        try:
            if not os.path.exists(path):
                _path, _file = os.path.split(path)
                if (len(_path) > 0) and (not os.path.exists(_path)):
                    os.makedirs(_path)
                with open(path, "w") as f:
                    f.write("")

            with open(path, 'a') as f:
                f.write('')
            AmpConfig.file_logging = path
            return 0
        except IOError as x:
            if x.errno == errno.EACCES:
                sys.stderr.write("Can't access the file: %s\n" % path)
            elif x.errno == errno.EISDIR:
                sys.stderr.write("This is a folder, a file expected: %s\n" % path)
            else:
                sys.stderr.write("Other error to create configuration file: %s" % str(x))
            return -1

    @staticmethod
    def random_alphanum(length, lowercase=False):
        """
        Generates a random password which consists of digits, lowercase and uppercase characters
        :param lowercase:
        :param length:
        :return:
        """
        if lowercase:
            return ''.join(random.SystemRandom().choice(string.ascii_lowercase + string.digits) for _ in range(length))
        else:
            return ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(length))

    @staticmethod
    def parse_params():

        logging.getLogger().setLevel(logging.INFO)  # default level
        AmpConfig.set_ca_env()  # this may change later
        just_send_ca = False

        # let's find out if the config folder is there
        if AmpConfig._test_writable(AmpConfig.config_dir) != 0:
            sys.stderr.write("Please check that %s is a directory and you have a write permissions.\n\n" % AmpConfig.config_dir)
            return -1

        raw_config = {}
        file_name = os.path.join(AmpConfig.config_dir, 'config.json')
        if os.path.exists(file_name):
            try:
                with open(file_name, 'r') as f:
                    raw_config = json.loads(f.read())
                    cur_id = raw_config['id']
            except Exception as ex:
                logging.error("The config file exists but I can't read it", cause=str(ex), file=file_name)
                return -1
        else:
            cur_id = AmpConfig.random_alphanum(24, lowercase=True) + "@amp.keychest.net"
            raw_config['id'] = cur_id
            try:
                with open(file_name, 'w') as f:
                    f.write(json.dumps(raw_config, indent=4))
            except Exception as ex:
                logging.error("The config file exists but I can't read it", cause=str(ex), file=file_name)
                return -1

        AmpConfig.id = cur_id  # setting the proxy ID

        try:
            opts, args = getopt.getopt(sys.argv[1:], 'hl:nf:cp:a:e:',
                                       ['help', 'level=', 'noapi', 'file=', 'ca', 'port=', 'addr=', 'etc='])
        except getopt.GetoptError:
            # print help information and exit:
            AmpConfig.print_help()
            sys.stderr.write("You provided %s\n" % ' '.join(sys.argv[1:]))
            if AmpConfig.ch:
                AmpConfig.ch.close()
            return -1

        for o, a in opts:
            if o in ('-h', '--help'):
                AmpConfig.print_help()
                if AmpConfig.ch:
                    AmpConfig.ch.close()
                return 1
            else:
                if o in ('-n', '--noapi'):
                    AmpConfig.api_logging = None
                    if AmpConfig.file_logging is None:
                        AmpConfig.file_logging = '/var/log/keychestamp/audit.json'
                if o in ('-l', '--level'):
                    if a.lower() in ['debug', 'info', 'warning', 'error', 'critical']:
                        if a.lower() == 'debug':
                            AmpConfig.log_level = logging.DEBUG
                        elif a.lower() == 'info':
                            AmpConfig.log_level = logging.INFO
                        elif a.lower() == 'warning':
                            AmpConfig.log_level = logging.WARNING
                        elif a.lower() == 'error':
                            AmpConfig.log_level = logging.ERROR
                        elif a.lower() == 'critical':
                            AmpConfig.log_level = logging.CRITICAL
                        logging.getLogger().setLevel(AmpConfig.log_level)
                if o in ('-e', '--etc'):
                    try:
                        AmpConfig.config_dir = a
                        if AmpConfig.config_dir and (AmpConfig.config_dir[len(AmpConfig.config_dir) - 1] == os.path.sep):
                            AmpConfig.config_dir = AmpConfig.config_dir[:-1]

                        if os.path.exists(a):
                            if os.path.isdir(a):
                                if AmpConfig._test_writable(AmpConfig.config_dir) != 0:
                                    return -1
                            else:
                                sys.stderr.write("Working directory is not a directory\n\n")
                                return -1
                        else:
                            os.makedirs(a)
                            if AmpConfig._test_writable(AmpConfig.config_dir) != 0:
                                return -1
                            pass
                        AmpConfig.set_ca_env()
                    except Exception as ex:
                        sys.stderr.write("Working directory is incorrect: %s\n" % str(ex))
                if o in ('-f', '--file'):
                    _x = AmpConfig._test_config(a)
                    if _x != 0:
                        return -1
                    AmpConfig.file_logging = a
                if o in ('-p', '--port'):
                    try:
                        new_port = int(a)
                        if (new_port < 1) or (new_port > 65535):
                            sys.stderr.write("The port is not a valid value between 1-65535\n\n")
                            return -1
                        AmpConfig.port = new_port
                    except ValueError:
                        sys.stderr.write("The port you entered is not an integer\n\n")
                        return -1
                if o in ('-a', '--addr'):
                    AmpConfig.address = a.strip()
                if o in ('-c', '--ca'):
                    just_send_ca = True

        sys.stderr.write("Config directory is %s\n" % AmpConfig.config_dir)

        AmpConfig.ensureCA()

        if just_send_ca:
            cert = None
            try:
                with open(AmpConfig.ca_cert, "r") as f:
                    cert = f.read()
            except Exception as ex:
                sys.stderr.write("Reading CA cert was not successful: %s\n\n" % str(ex))
                pass
            if cert:
                sys.stdout.write(cert)
                sys.stdout.flush()
            return 1
        else:
            if AmpConfig.file_logging is not None:
                _x = AmpConfig._test_config(AmpConfig.file_logging)
                if _x != 0:
                    return -1
            return 0  # all ok

    @staticmethod
    def set_logging():
        # noinspection PyUnusedLocal
        def usage_log(message, *args, **kws):
            # noinspection PyUnresolvedReferences
            if logging.getLogger().isEnabledFor(logging.OPS_LEVEL):
                # let's do JSON encoding here
                # noinspection PyUnresolvedReferences
                message = QueueHandler.create_json_msg(message, **kws)
                # noinspection PyUnresolvedReferences
                logging.log(logging.OPS_LEVEL, message)

        # noinspection PyUnusedLocal
        def debug_log(message, *args, **kws):
            if logging.getLogger().isEnabledFor(logging.DEBUG):
                message = QueueHandler.create_json_msg(message, **kws)
                logging.log(logging.DEBUG, message)

        # noinspection PyUnusedLocal
        def info_log(message, *args, **kws):
            if logging.getLogger().isEnabledFor(logging.INFO):
                message = QueueHandler.create_json_msg(message, **kws)
                logging.log(logging.INFO, message)

        # noinspection PyUnusedLocal
        def warning_log(message, *args, **kws):
            if logging.getLogger().isEnabledFor(logging.WARNING):
                message = QueueHandler.create_json_msg(message, **kws)
                logging.log(logging.WARNING, message)

        # noinspection PyUnusedLocal
        def error_log(message, *args, **kws):
            if logging.getLogger().isEnabledFor(logging.ERROR):
                message = QueueHandler.create_json_msg(message, **kws)
                logging.log(logging.ERROR, message)

        # noinspection PyUnusedLocal
        def critical_log(message, *args, **kws):
            if logging.getLogger().isEnabledFor(logging.CRITICAL):
                message = QueueHandler.create_json_msg(message, **kws)
                logging.log(logging.CRITICAL, message)

        logging.OPS_LEVEL = 60
        # noinspection PyUnresolvedReferences
        logging.addLevelName(logging.OPS_LEVEL, "USE")
        logging.Logger.usage = usage_log
        logging.usage = usage_log

        logging.Logger.debug = debug_log
        logging.debug = debug_log
        logging.Logger.warning = warning_log
        logging.warning = warning_log
        logging.Logger.warn = warning_log
        logging.warn = warning_log
        logging.Logger.info = info_log
        logging.info = info_log
        logging.Logger.error = error_log
        logging.error = error_log
        logging.Logger.critical = critical_log
        logging.critical = critical_log

        # items to be added to all usage logs
        logging.use_counter = 1
        logging.host = socket.gethostname()
        logging.ipaddress = socket.gethostbyname(socket.gethostname())
        logging.proxy_id = AmpConfig.id

        logger = logging.getLogger()
        logging.basicConfig(datefmt='%Y-%m-%d %H:%M:%S')
        logger.setLevel(AmpConfig.log_level)
        AmpConfig.ch = QueueHandler(api_logging=AmpConfig.api_logging, file_logging=AmpConfig.file_logging)
        fmt = logging.Formatter('%(asctime)s:%(name)s:%(levelname)s: %(message)s')
        AmpConfig.ch.setFormatter(fmt)
        AmpConfig.ch.setLevel(logging.INFO)
        for x in logger.handlers:
            logger.removeHandler(x)
        logger.addHandler(AmpConfig.ch)
        pass

    @staticmethod
    def join_with_script_dir(path):
        return os.path.join(os.path.dirname(os.path.abspath(__file__)), path)

    @staticmethod
    def ensure_name_cert(hostname):

        cert_path = "%s/%s.crt" % (AmpConfig.cert_dir.rstrip('/'), hostname)
        AmpConfig.update_pki_config(AmpConfig.config, hostname)
        if not os.path.isfile(cert_path):
            epoch = "%d" % (time.time() * 1000)
            # openssl 1.1. allows req with   "-addext", "subjectAltName = DNS:%s" % hostname
            p1 = Popen(["openssl", "req", "-new", '-inform', 'pem', "-nodes", "-key", AmpConfig.cert_key, "-extensions", "server_cert",
                        "-subj", "/CN=%s" % hostname, '-config', AmpConfig.config],
                       stdout=PIPE)

            p2 = Popen(["openssl", "x509", "-req", "-days", "365", "-CA", AmpConfig.ca_cert, "-CAkey", AmpConfig.ca_key,
                        "-set_serial", epoch, "-sha256", "-extfile", AmpConfig.config, '-extensions', 'server_cert', "-out", cert_path],
                       stdin=p1.stdout, stderr=PIPE)
            p2.communicate()
        return None

    @staticmethod
    def update_pki_config(_path, domain=None):
        # create the configuration file for the CA
        with open(_path, 'w') as f:
            f.writelines([
                '[ ca ]\n',
                'default_ca = CA_default\n',
                '\n',
                '[ CA_default ]\n',
                'dir               = %s\n' % AmpConfig.ca_dir,
                'certs             = $dir/certs\n',
                'crl_dir           = $dir/crl\n',
                'new_certs_dir     = $dir/newcerts\n',
                'database          = $dir/index.txt\n',
                'serial            = $dir/serial\n',
                'default_days      = 3650\n',
                'RANDFILE          = $dir/private/.rand\n',
                'private_key       = $dir/private/ca.key\n',
                'certificate       = $dir/certs/ca.crt\n',
                'default_md        = sha256\n',
                'preserve          = no\n',
                'name_opt          = ca_default\n',
                'cert_opt          = ca_default\n',
                'policy            = policy_strict\n',
                'copy_extensions   = copy\n',
                '\n',
                '[ req ]\n',
                'default_bits            = 2048\n',
                'default_md              = sha256\n',
                'distinguished_name      = req_dn\n',
                'prompt                  = yes\n',
                'encrypt_key             = no\n',
                'string_mask             = utf8only\n',
                '\n',
                '[ v3_ca ]\n',
                'basicConstraints        = critical,CA:true\n',
                'subjectKeyIdentifier    = hash\n',
                'authorityKeyIdentifier = keyid:always,issuer\n',
                'keyUsage                = critical,keyCertSign,cRLSign,digitalSignature\n',
                '\n',
                '[ req_dn ]\n',
                'countryName                     = Country Name (2 letter code)\n',
                '0.organizationName              = Organization Name\n',
                'organizationalUnitName          = Organizational Unit Name\n',
                'commonName                      = Common Name\n',
                '\n',
                'countryName_default             = US\n',
                '0.organizationName_default        = KeyChest\n',
                'organizationalUnitName_default  = KeyChestAmp\n',
                '\n',
                '[ policy_loose ]\n',
                'countryName             = optional\n',
                'organizationName        = optional\n',
                'organizationalUnitName  = optional\n',
                'commonName              = supplied\n',
                '\n',
                '[ policy_strict ]\n,'
                'countryName             = match\n',
                'organizationName        = match\n',
                'organizationalUnitName  = optional\n',
                'commonName              = supplied\n',
                '\n',
                '[ server_cert ]\n',
                'basicConstraints       = CA:FALSE\n',
                'subjectKeyIdentifier   = hash\n',
                'authorityKeyIdentifier = keyid,issuer:always\n',
                'keyUsage               = critical,digitalSignature,keyEncipherment\n',
                'extendedKeyUsage       = serverAuth\n'
            ])
            if domain:
                f.writelines([
                    'subjectAltName=DNS:%s\n' % domain
                ])

    @staticmethod
    def ensureCA():
        if not os.path.exists(AmpConfig.ca_dir):
            os.mkdir(AmpConfig.ca_dir, 0o770)
        if not os.path.exists(AmpConfig.ca_dir + '/index.txt'):
            with open(AmpConfig.ca_dir + '/index.txt', 'w') as f:
                f.write("")
        if not os.path.exists(AmpConfig.ca_dir + '/serial'):
            with open(AmpConfig.ca_dir + '/serial', 'w') as f:
                f.write("1000")
        if not os.path.exists(AmpConfig.ca_dir + '/certs'):
            os.mkdir(AmpConfig.ca_dir + '/certs', 0o770)
        if not os.path.exists(AmpConfig.ca_dir + '/crl'):
            os.mkdir(AmpConfig.ca_dir + '/crl', 0o770)
        if not os.path.exists(AmpConfig.ca_dir + '/newcerts'):
            os.mkdir(AmpConfig.ca_dir + '/newcerts', 0o770)
        if not os.path.exists(AmpConfig.ca_dir + '/private'):
            os.mkdir(AmpConfig.ca_dir + '/private', 0o700)

        AmpConfig.update_pki_config(AmpConfig.config)

        if (not os.path.isfile(AmpConfig.ca_key)) or (not os.path.isfile(AmpConfig.ca_cert)) or (not os.path.isfile(AmpConfig.cert_key)):
            # and os.path.isfile(certkey) and os.path.isdir(certdir):
            # generate a new CA key
            p1 = Popen(["openssl", "genrsa", "-out", AmpConfig.ca_key], stderr=PIPE)
            # create a new CA cert
            p1.communicate()
            p2 = Popen(["openssl", "req", "-new", "-x509", '-inform', 'pem', "-nodes", "-key", AmpConfig.ca_key,
                        "-extensions", "v3_ca", "-out", AmpConfig.ca_cert, "-days", "3650", "-config", AmpConfig.config,
                        '-subj', '/C=US/O=KeyChest/OU=KeyChest ACMEv2 Proxy'],
                       stdout=PIPE, stderr=PIPE)
            p2.communicate()
            p3 = Popen(['openssl', 'genrsa', '-out', AmpConfig.cert_key, "2048"],
                       stderr=PIPE)
            p3.communicate()
        pass

    @staticmethod
    def cert_ready():
        return os.path.isfile(AmpConfig.ca_key) and os.path.isfile(AmpConfig.ca_cert) \
               and os.path.isfile(AmpConfig.cert_key) and os.path.isdir(AmpConfig.cert_dir)
        pass

    @staticmethod
    def get_cert_path(hostname):

        _path = "%s/%s.crt" % (AmpConfig.cert_dir.rstrip('/'), hostname)
        return _path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

https://mattgathu.github.io/multiprocessing-logging-in-python/

 Copyright (C) KeyChest Ltd, registered in the United Kingdom.
 This file is owned exclusively by KeyChest Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import datetime
import json
import time
import requests

__author__ = "Dan Cvrcek"
__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'

import sys
import logging
import traceback
import threading
import multiprocessing
# noinspection PyCompatibility
from queue import Empty  # future is required by setup for py3
# noinspection PyPep8Naming
from logging import FileHandler as FH

from keychestamp.version import proxy_version


# ============================================================================
# Define Log Handler
# ============================================================================
# noinspection PyMissingOrEmptyDocstring
class QueueHandler(logging.Handler):
    """multiprocessing log handler

    This handler makes it possible for several processes
    to log to the same file by using a queue.

    """
    def __init__(self, api_logging=True, file_logging=None, fname=None):
        logging.Handler.__init__(self)

        if fname:
            self._handler = FH(fname)
        else:
            self._handler = logging.StreamHandler(sys.stderr)

        self.queue = multiprocessing.Queue(-1)
        self.queue_usage = multiprocessing.Queue(-1)
        self.queue_lock = threading.Lock()

        self.queue_usage_len = 0
        self.queue_len = 0
        self.usage_file = file_logging
        self.file_handle = None
        if api_logging:
            self.usage_url = "https://keychest.net/api/v1.0/amp/event"
        else:
            self.usage_url = None

        self.headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
        self.stop = multiprocessing.Event()

        thrd = threading.Thread(target=self.receive, args=(self.stop,), name="log listener")
        thrd.daemon = True
        thrd.start()
        thrd2 = threading.Thread(target=self.send_usage, args=(self.stop,), name="log pusher")
        thrd2.daemon = True
        thrd2.start()

    def setUsageFile(self, path):
        """

        :param path:
        :return:
        """
        pass

    def setUsageURL(self, path):
        """

        :param path:
        :return:
        """
        pass

    def setFormatter(self, fmt):
        logging.Handler.setFormatter(self, fmt)
        self._handler.setFormatter(fmt)

    def setLevel(self, level):
        self._handler.setLevel(level)

    def send_usage(self, stop_event):
        """

        :param stop_event
        :type stop_event: multiprocessing.Event
        :rtype: None
        """
        while not stop_event.is_set():  # type: multiprocessing.Event
            try:
                record = self.queue_usage.get(timeout=0.5)
                if record is None:
                    continue

                with self.queue_lock:
                    self.queue_len -= 1

                if self.usage_url or self.usage_file:
                    j_record = json.loads(record)
                    j_record['backlog'] = self.queue_len
                    j_record['log_time'] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

                if self.usage_file:
                    if (self.file_handle is None) or self.file_handle.closed:
                        self.file_handle = open(self.usage_file, 'a')
                    # noinspection PyUnboundLocalVariable
                    self.file_handle.write(json.dumps(j_record)+"\n")
                    pass
                if self.usage_url is not None:
                    code = 0
                    while (code != 200) and (not stop_event.is_set()):
                        r = requests.post(url=self.usage_url, data=j_record, headers=self.headers, timeout=5)
                        code = r.status_code
                        if code != 200:
                            time.sleep(5)
                        # todo gracious treatment of the API down - create a temp file and store the data there till the API is ready

                    pass
            except Empty:
                if self.file_handle:
                    self.file_handle.flush()
                pass  # empty queue
            except (KeyboardInterrupt, SystemExit):
                if self.file_handle:
                    self.file_handle.flush()
                raise
            except EOFError:
                if self.file_handle:
                    self.file_handle.flush()
                break
            except BaseException as ex:
                if self.file_handle:
                    self.file_handle.flush()
                traceback.print_exc(file=sys.stderr)
                sys.stderr.write("Exception in queue_handler %s" % str(ex))
        sys.stderr.write("queue handler usage terminated\n")

    def receive(self, stop_event):
        """

        :param stop_event
        :type stop_event: multiprocessing.Event
        :rtype: object
        """
        while not stop_event.is_set():  # type: multiprocessing.Event
            try:
                record = self.queue.get(timeout=0.5)
                if record is not None:
                    self._handler.emit(record)
            except Empty:
                pass  # empty queue
            except (KeyboardInterrupt, SystemExit):
                raise
            except EOFError:
                break
            except BaseException as ex:
                traceback.print_exc(file=sys.stderr)
                sys.stderr.write("Exception in queue_handler %s" % str(ex))
        sys.stderr.write("queue handler logs terminated\n")

    def send(self, s):
        self.queue.put_nowait(s)

    def _format_record(self, record):
        if record.args:
            record.msg = record.msg % record.args
            record.args = None
        if record.exc_info:
            # noinspection PyUnusedLocal
            dummy = self.format(record)
            record.exc_info = None

        return record

    def emit(self, record):
        """

        :param self:
        :param record:
        """
        # noinspection PyBroadException
        try:
            if record.levelname == 'USE':
                self.queue_usage.put_nowait(record.msg)
                with self.queue_lock:
                    self.queue_len += 1
                pass
            else:
                s = self._format_record(record)
                self.send(s)
        except (KeyboardInterrupt, SystemExit):
            raise
        except BaseException:
            self.handleError(record)

    @staticmethod
    def create_json_msg(message, **kws):
        # noinspection PyUnresolvedReferences
        _temp = {
            'msg': message,
            'params': {},
            'id': logging.use_counter,
            'proxy': logging.proxy_id,
            'time': datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            'host': logging.host,
            'ip': logging.ipaddress,
            'version': proxy_version
        }
        # noinspection PyUnresolvedReferences
        logging.use_counter += 1
        if kws and (len(kws) > 0):
            for key, item in kws.items():
                # if key.lower() == 'code':
                #     _temp['code'] = item
                # else:
                #     _temp['params'][key] = item
                _temp['params'][key] = item

        message = json.dumps(_temp)
        return message

    def close(self):
        self.stop.set()
        if self._handler:
            self._handler.close()
        if logging.Handler:
            logging.Handler.close(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module:
***

 Copyright (C) KeyChest Ltd, registered in the United Kingdom.
 This file is owned exclusively by KeyChest Ltd.
 Unauthorized copying of this file, via any medium is strictly prohibited
 Proprietary and confidential
 Written by Dan Cvrcek <support@keychest.net>, 2019
"""
import hashlib
import json
import os
import pickle
import platform
# import resource
import sys
import threading
import time
import traceback
import socket
from multiprocessing import Queue
from threading import Lock

from logbook import NOTSET
from logbook.handlers import Handler
import logbook


__copyright__ = 'KeyChest Ltd'
__email__ = 'support@keychest.net'
__status__ = 'Development'


# noinspection PyPep8Naming,PyBroadException
class logger(object):
    """
    The available logging levels are
    CRITICAL = 15  ... we use this for "usage" so we can trace actual database updates
    ERROR = 14  – for errors that occur, but are handled
    WARNING = 13  – for exceptional circumstances that might not be errors
    NOTICE = 12 – for non-error messages you usually want to see
    INFO = 11 – for messages you usually don’t want to see
    DEBUG = 10 – for debug messages
    TRACE = 9 - temporary
    """

    log_codes = {}  # a dictionary, values are tuples of int:code and str:message
    filename = None
    last_code = 0  # we will always start from 0
    code_gap = []
    file_lock = threading.Lock()
    msg_id = 1

    # create a mapping table -
    @classmethod
    def load_codes(cls, filename=None):
        """
        Load error message codes from backup file
        :param filename:
        :return:
        """
        if cls.filename is not None:
            return  # repeated call
        if filename:
            cls.filename = filename
        if cls.filename and os.path.isfile(cls.filename):
            file_cache = open(cls.filename, "rb")
            try:
                cls.log_codes = pickle.load(file_cache)
            except Exception:
                # file format is bad - fix it
                cls.log_codes = {}
                os.rename(cls.filename, cls.filename + '.corrupted %d' % int(time.time()))
                cls.update_codes()

            # lets' get all the values
            existing_codes = [a[0] for a in list(cls.log_codes.values())]
            existing_codes.sort()
            cls.code_gap = []
            if len(existing_codes) > 0:  # and find gaps we can re-use
                cls.last_code = existing_codes[-1]
                # store unused code values
                for i in range(1, cls.last_code + 1):
                    if i not in existing_codes:
                        cls.code_gap.append(i)
            else:
                pass
        elif cls.filename:
            directory = os.path.dirname(cls.filename)
            os.makedirs(directory, exist_ok=True)
            open(cls.filename, 'a').close()
            cls.log_codes = {}
            cls.update_codes()  # create a pickle file for empty dictionary
        else:
            cls.log_codes = {}
            pass

    @classmethod
    def update_codes(cls):
        """
        Update error codes
        :return:
        """
        if cls.filename is None:
            return
        file_cache = open(cls.filename+"_new", "wb")
        pickle.dump(cls.log_codes, file_cache)
        file_cache.close()
        os.remove(cls.filename)
        os.rename(cls.filename+"_new", cls.filename)

    @classmethod
    def get_code(cls, level, key, msg=None):
        """
        Get a code for a given message
        :param level:
        :param key:
        :param msg:
        :return:
        """

        key = None
        if key:
            if key in cls.log_codes:
                code = cls.log_codes[key][0]
            else:
                if len(cls.code_gap) > 0:
                    code = cls.code_gap.pop(0)
                    cls.log_codes[key] = (code, msg)
                else:
                    while cls.last_code + 1 in cls.log_codes:
                        cls.last_code += 1
                    cls.last_code += 1
                    cls.log_codes[key] = (cls.last_code, msg)
                    code = cls.last_code

                cls.file_lock.acquire()
                try:
                    cls.update_codes()
                finally:
                    cls.file_lock.release()
            result = "%s%04d" % (level[:3], code)
        else:
            code = "____"
            result = "%s%s" % (level[:3], code)

        if msg is None:
            try:
                raise Exception
            except Exception:
                tb = "; ".join(traceback.extract_stack().format())
                logger.error('Null message passed to logger',
                             name=threading.current_thread().name, traceback=tb)
        cls.file_lock.acquire()
        _id = cls.msg_id  # counter of log messages generated during this run - a simple counter
        cls.msg_id += 1
        cls.file_lock.release()
        return result, _id

    @classmethod
    def get_hash(cls, args):
        """
        Compute a hash value for the first of the arguments
        :param args:
        :return:
        """
        if len(args) > 0:
            return hashlib.sha1(args[0].encode()).hexdigest()
        else:
            return None

    @classmethod
    def get_msg_details(cls, msg, dict_params):
        """
        Get details for submitting a log message.
        :param msg:
        :param dict_params:
        :return:
        """
        key = cls.get_hash(msg)

        params = "{}"
        try:
            params = json.dumps(dict_params)
        except Exception:
            sys.stderr.out("Can't parse this %s %s\n" % (msg[0], dict_params))
            pass
        finally:
            if dict_params is not None and 'traceback' in dict_params:
                del dict_params['traceback']

        return key, params

    @classmethod
    def trace(cls, *args, **kwargs):
        """
        Trace level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.trace(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def debug(cls, *args, **kwargs):
        """
        debug level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.debug(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def info(cls, *args, **kwargs):
        """
        Info level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.info(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def notice(cls, *args, **kwargs):
        """
        notice level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.notice(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def warn(cls, *args, **kwargs):
        """
        warning level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.warn(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def warning(cls, *args, **kwargs):
        """
        warning level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.warning(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def error(cls, *args, **kwargs):
        """
        error level message
        :param args:
        :param kwargs:
        :return:
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.error(*args, code=key, usage=False, params=params)
        except Exception:
            pass

    @classmethod
    def usage(cls, *args, **kwargs):
        """
        We use "critical" so it's always logged!
        :param args:
        :param kwargs:
        :return: None
        """
        key, params = cls.get_msg_details(args, kwargs)
        try:
            logbook.critical(*args, code=key, usage=True, params=params)
        except Exception:
            pass


class TCPHandler(Handler):
    """A custom handler using sock file or TCP sockets on a given port. It's been created from Redis Handler

    It publishes each record as json dump.

    Example setup::

        handler = TCPHandler('http://127.0.0.1', port='9200', key='redis')

    If your Redis instance is password protected, you can securely connect
    passing your password when creating a RedisHandler object.

    Example::

        handler = RedisHandler(password='your_redis_password')

    More info about the default buffer size: wp.me/p3tYJu-3b
    """
    def __init__(self, sock='logbook.sock', ip=None, port=None,
                 extra_fields=None, flush_threshold=128, flush_time=1,
                 level=NOTSET, _filter=None, bubble=True):
        Handler.__init__(self, level, _filter, bubble)

        if ip is None:
            self.sock_type = socket.AF_UNIX
            self.sock_address = sock

        else:
            self.sock_type = socket.AF_INET
            self.sock_address = (ip, port)

        self.sock = socket.socket(self.sock_type)
        self.log_counter = 0

        connected = False
        while not connected:
            try:
                self.sock.connect(self.sock_address)
                connected = True
                self.sock.close()
            except Exception as ex:
                sys.stderr.write("INFO: Logging server not available (%s)\n" % str(ex))
                sys.stderr.flush()
                time.sleep(2)

        self.extra_fields = extra_fields or {}
        self.flush_threshold = flush_threshold
        self.queue = Queue()
        self.last_error_time = time.time()
        self.lock = Lock()

        # Set up a thread that flushes the queue every specified seconds
        self._stop_event = threading.Event()
        self._flushing_t = threading.Thread(target=self._flush_task, args=(flush_time, True))
        self._flushing_t.daemon = True
        self._flushing_t.name = "Log flush"
        # self._flush_buffer()
        self._flushing_t.start()

    # noinspection PyUnusedLocal
    def _flush_task(self, wait_time, dummy):
        """Calls the method _flush_buffer every certain time.
        """
        while not self._stop_event.isSet():
            with self.lock:
                self._flush_buffer()
            self._stop_event.wait(wait_time)

    def _flush_buffer(self):
        """Flushes the messaging queue into Redis.

        All values are pushed at once for the same key.

        The method rpush/lpush is defined by push_method argument
        """
        if self.queue:
            new_socket = socket.socket(self.sock_type)
            try:
                connected = False
                while not self.queue.empty():
                    message = self.queue.get()
                    message += os.linesep
                    if not connected:
                        new_socket.connect(self.sock_address)
                        connected = True
                    new_socket.sendall(message.encode())
            except Exception as ex:
                if time.time() - self.last_error_time > 1:
                    self.last_error_time = time.time()
                    sys.stderr.write("Error connecting to log dispatcher %s\n" % str(ex))
                    sys.stderr.flush()
            finally:
                # noinspection PyBroadException
                try:
                    if new_socket is not None:
                        new_socket.close()
                except Exception:
                    pass

    def disable_buffering(self):
        """Disables buffering.

        If called, every single message will be directly pushed to Redis.
        """
        self._stop_event.set()
        self.flush_threshold = 1

    def emit(self, record):
        """Emits a pair (key, value) to redis.

        The key is the one provided when creating the handler, or redis if none
        was provided. The value contains both the message and the hostname.
        Extra values are also appended to the message.
        """
        # print("XXXXX emit")
        with self.lock:
            r = {"message": record.msg,
                 "host": platform.node(),
                 "level": record.level_name,
                 "time": record.time.isoformat()}
            r.update(self.extra_fields)
            r.update(record.kwargs)
            self.queue.put(json.dumps(r))
            self.log_counter += 1
            if self.log_counter >= self.flush_threshold:
                self.log_counter = 0
                self._flush_buffer()

    def close(self):
        """
        Flush the log bugger
        :return:
        """
        self._flush_buffer()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
***
Module: keychestamp
***
"""
# Copyright (C) KeyChest Ltd, registered in the United Kingdom.
#
# This file is provided under a license as specified in the "LICENSE" file, which is part
# of this software package.
#
# Written by KeyChest <support@keychest.net>, 2019

import sys
from setuptools import setup, find_packages
from keychestamp.version import proxy_version

version = proxy_version

# pypi publishing
# 1. set $HOME/.pypirc
#      [distutils]
#      index-servers =
#          pypi
#
#      [pypi]
#      username: <name>
#      password: <password>
# 2. deactivate  // if there's an active env
# 3. cd pycharmenv3; source bin/activate
# 4. pip3 install --upgrade wheel setuptools twine
# 5. cd <whatever_to>/keychestamp
# 6. rm -rf dist/*
# 7. python3 setup.py sdist bdist_wheel
# 7a.twine check dist/*
# 8. twine upload dist/<latest>.tar.gz
# 9. you can test it with "pip3 install --upgrade --no-cache-dir keychestamp"


install_requires = [
    'requests',
    'certifi'
]

if sys.version_info < (3,):
    print("Python 2 and Python 3.5 or lower are not supported")
    pass

setup(
    name='keychestamp',
    version=version,
    packages=find_packages(),
    namespace_packages=['keychestamp'],
    include_package_data=True,
    package_data={'keychestamp': ['LICENSE']},
    install_requires=install_requires,
    url='https://keychest.net',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    author='KeyChest Ltd',
    author_email='support@keychest.net',
    description='ACMEv2 proxy to manage clients and observe rate limits of Let\'s Encrypt',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: System :: Networking',
        'Topic :: Internet :: Proxy Servers'
    ],
    extras_require={
        # 'dev': dev_extras,
        # 'docs': docs_extras,
    },

    entry_points={
        'console_scripts': [
            'keychestamp=keychestamp.main:main'
        ],
    }
)
